;; Generated package description from ffs.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "ffs" "0.2.3snapshot0.20260522.1" "Form feed-separated plain text presentations" '((emacs "27.1")) :commit "9408e387f38c4dcd71f378b6ca7c2ee9cca8b0d7" :authors '(("Amin Bandali" . "bandali@gnu.org")) :maintainer '("Amin Bandali" . "bandali@gnu.org") :keywords '("convenience" "focus" "narrowing" "outlines" "presentation" "text") :url "https://git.kelar.org/~bandali/ffs")
